
/////////////////////////////////////////////////////////////////////////
//
// CD-TEXT decoding routine (CdText.cpp)
//
//   Copyright (C) 2002 Takabin (Bin Takanami)
//   This file is NOT under GPL, you can use it freely at your own risk.
//
/////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "CdText.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif
//---------------------------------------------------------------------------

bool ReadEachLang(TCdTextPacket* inTextPackets, int inNumPackets,
    TCdTextLangBlock *outLangBlock);

void SeparateInfo(char *inBuffer, int inBufSize, void* outTracks,
    int inNumTracks, bool inIs2byteChar);

bool Is2ByteCharset(CdTextCharsets inCharset);
//---------------------------------------------------------------------------

// CD-TEXTfR[hEGW
//
//
bool DecodeCdText(
    TCdTextPacket* inTextPackets,
    int inNumPackets,
    CdTextDisc *outTextInfo)
{

    memset(outTextInfo, 0, sizeof(CdTextDisc));

    if(inNumPackets < 4) return false;

    TCdTextPacket* pPacket;

    SizeInfo1* pSizeInfo1 = NULL;
    SizeInfo2* pSizeInfo2 = NULL;
    SizeInfo3* pSizeInfo3 = NULL;

    pPacket = inTextPackets + (inNumPackets - 3);
    if((pPacket->mType == ctSizeInfo) && (pPacket->mTrackNo == 0))
        pSizeInfo1 = (SizeInfo1*)pPacket->mData;

    pPacket = inTextPackets + (inNumPackets - 2);
    if((pPacket->mType == ctSizeInfo) && (pPacket->mTrackNo == 1))
        pSizeInfo2 = (SizeInfo2*)pPacket->mData;

    pPacket = inTextPackets + (inNumPackets - 1);
    if((pPacket->mType == ctSizeInfo) && (pPacket->mTrackNo == 2))
        pSizeInfo3 = (SizeInfo3*)pPacket->mData;

    if(!pSizeInfo1 || !pSizeInfo2 || !pSizeInfo3) {
        return false;
    }

    int i;

    int theNumPackets[MAX_NUM_LANG_BLOCKS];
    memset(theNumPackets, 0, sizeof(int) * MAX_NUM_LANG_BLOCKS);

    for(i=0;i<MAX_NUM_LANG_BLOCKS;i++){
        switch(i){
        case 0:
            theNumPackets[i] = pSizeInfo2->mNumPackets1 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage1;
            break;
        case 1:
            theNumPackets[i] = pSizeInfo2->mNumPackets2 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage2;
            break;
        case 2:
            theNumPackets[i] = pSizeInfo2->mNumPackets3 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage3;
            break;
        case 3:
            theNumPackets[i] = pSizeInfo2->mNumPackets4 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage4;
            break;
        case 4:
            theNumPackets[i] = pSizeInfo3->mNumPackets5 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage5;
            break;
        case 5:
            theNumPackets[i] = pSizeInfo3->mNumPackets6 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage6;
            break;
        case 6:
            theNumPackets[i] = pSizeInfo3->mNumPackets7 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage7;
            break;
        case 7:
            theNumPackets[i] = pSizeInfo3->mNumPackets8 + 1;
            outTextInfo->mLangBlocks[i].mCharset  = (CdTextCharsets) pSizeInfo1->mCharset;
            outTextInfo->mLangBlocks[i].mLanguage = (CdTextLanguages)pSizeInfo3->mLangage8;
            break;
        }
    }

    int theNumLangBlocks = MAX_NUM_LANG_BLOCKS;
    for(i=0;i<MAX_NUM_LANG_BLOCKS;i++){
        if(theNumPackets[i] == 0){
            theNumLangBlocks = i;
            break;
        }
    }
    outTextInfo->mNumLangBlocks = theNumLangBlocks;

    int theNumOffsetPackets = 0;
    for(i=0;i<theNumLangBlocks;i++){

        if(!ReadEachLang(inTextPackets + theNumOffsetPackets, theNumPackets[i],
                &outTextInfo->mLangBlocks[i])){
            outTextInfo->mNumLangBlocks = i;
            return true;
        }
        theNumOffsetPackets += theNumPackets[i];
    }

    return true;
}
//---------------------------------------------------------------------------

//eǂݏo
bool ReadEachLang(
    TCdTextPacket* inTextPackets,
    int inNumPackets,
    TCdTextLangBlock *outLangBlock)
{
    if(inNumPackets < 4) return false;

    TCdTextPacket* pPacket;

    SizeInfo1* pSizeInfo1 = NULL;
    SizeInfo2* pSizeInfo2 = NULL;
    SizeInfo3* pSizeInfo3 = NULL;

    pPacket = inTextPackets + (inNumPackets - 3);
    if((pPacket->mType == ctSizeInfo) && (pPacket->mTrackNo == 0))
        pSizeInfo1 = (SizeInfo1*)pPacket->mData;

    pPacket = inTextPackets + (inNumPackets - 2);
    if((pPacket->mType == ctSizeInfo) && (pPacket->mTrackNo == 1))
        pSizeInfo2 = (SizeInfo2*)pPacket->mData;

    pPacket = inTextPackets + (inNumPackets - 1);
    if((pPacket->mType == ctSizeInfo) && (pPacket->mTrackNo == 2))
        pSizeInfo3 = (SizeInfo3*)pPacket->mData;

    if(!pSizeInfo1 || !pSizeInfo2 || !pSizeInfo3) return false;

    char* pTitle      = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mTitle);
    char* pPerformer  = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mPerformer);
    char* pSongwriter = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mSongwriter);
    char* pComposer   = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mComposer);
    char* pArranger   = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mArranger);
    char* pMessage    = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mMessage);

    char* pGenre      = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mGenre);
    char* pIdentifier = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo1->mIdentifier);

    char* pISRC       = (char*)calloc(BLOCK_DATA_SIZE, pSizeInfo2->mMCNorISRC);

    bool theIs2byte = Is2ByteCharset((CdTextCharsets)pSizeInfo1->mCharset);

    int theCountG = 0;
    int theCountI = 0;

    int theCountT = 0;
    int theCountP = 0;
    int theCountS = 0;
    int theCountC = 0;
    int theCountA = 0;
    int theCountM = 0;

    int theCountMI = 0;

    int i;
    for(i=0;i<inNumPackets;i++){
        pPacket = inTextPackets + i;

        if(pPacket->mIs2ByteChar)
            theIs2byte=true;

        switch(pPacket->mType){
        case ctTitle:
            if(theCountT >= pSizeInfo1->mTitle) break;
            memcpy(pTitle + BLOCK_DATA_SIZE * theCountT, pPacket->mData, BLOCK_DATA_SIZE);
            theCountT++;
            break;
        case ctPerformer:
            if(theCountP >= pSizeInfo1->mPerformer) break;
            memcpy(pPerformer + BLOCK_DATA_SIZE * theCountP, pPacket->mData, BLOCK_DATA_SIZE);
            theCountP++;
            break;
        case ctSongwriter:
            if(theCountS >= pSizeInfo1->mSongwriter) break;
            memcpy(pSongwriter + BLOCK_DATA_SIZE * theCountS, pPacket->mData, BLOCK_DATA_SIZE);
            theCountS++;
            break;
        case ctComposer:
            if(theCountC >= pSizeInfo1->mComposer) break;
            memcpy(pComposer + BLOCK_DATA_SIZE * theCountC, pPacket->mData, BLOCK_DATA_SIZE);
            theCountC++;
            break;
        case ctArranger:
            if(theCountA >= pSizeInfo1->mArranger) break;
            memcpy(pArranger + BLOCK_DATA_SIZE * theCountA, pPacket->mData, BLOCK_DATA_SIZE);
            theCountA++;
            break;
        case ctMessage:
            if(theCountM >= pSizeInfo1->mMessage) break;
            memcpy(pMessage + BLOCK_DATA_SIZE * theCountM, pPacket->mData, BLOCK_DATA_SIZE);
            theCountM++;
            break;
        case ctIdentifier:
            if(theCountI >= pSizeInfo1->mIdentifier) break;
            memcpy(pIdentifier + BLOCK_DATA_SIZE * theCountI, pPacket->mData, BLOCK_DATA_SIZE);
            theCountI++;
            break;
        case ctGenre:
            if(theCountG >= pSizeInfo1->mGenre) break;
            memcpy(pGenre + BLOCK_DATA_SIZE * theCountG, pPacket->mData, BLOCK_DATA_SIZE);
            theCountG++;
            break;
        case ctMCNorISRC:
            if(theCountMI >= pSizeInfo2->mMCNorISRC) break;
            memcpy(pISRC + BLOCK_DATA_SIZE * theCountMI, pPacket->mData, BLOCK_DATA_SIZE);
            theCountMI++;
            break;
        }
    }

    if(pSizeInfo1->mGenre){
        outLangBlock->mGenre = ((int)pGenre[0])<< 8 | (int)pGenre[1];
        if(outLangBlock->mGenre){

            memcpy(outLangBlock->mGenreStr, pGenre + 2,
                min(BLOCK_DATA_SIZE * pSizeInfo1->mGenre - 2, MAX_TEXT - 2));
        }
    }

    if(pSizeInfo1->mIdentifier){
        memcpy(outLangBlock->mIndentifier, pIdentifier,
            min(BLOCK_DATA_SIZE * pSizeInfo1->mIdentifier, MAX_TEXT - 1));
    }

    outLangBlock->mFirstTrack = pSizeInfo1->mFirstTrack;
    outLangBlock->mLastTrack  = pSizeInfo1->mLastTrack;

    int theNumTracks = pSizeInfo1->mLastTrack - pSizeInfo1->mFirstTrack + 1;

    SeparateInfo(pTitle, BLOCK_DATA_SIZE * pSizeInfo1->mTitle,
        outLangBlock->mTitle, theNumTracks, theIs2byte);
    SeparateInfo(pPerformer, BLOCK_DATA_SIZE * pSizeInfo1->mPerformer,
        outLangBlock->mPerformer, theNumTracks, theIs2byte);
    SeparateInfo(pSongwriter, BLOCK_DATA_SIZE * pSizeInfo1->mSongwriter,
        outLangBlock->mSongwriter, theNumTracks, theIs2byte);
    SeparateInfo(pComposer, BLOCK_DATA_SIZE * pSizeInfo1->mComposer,
        outLangBlock->mComposer, theNumTracks, theIs2byte);
    SeparateInfo(pArranger, BLOCK_DATA_SIZE * pSizeInfo1->mArranger,
        outLangBlock->mArranger, theNumTracks, theIs2byte);
    SeparateInfo(pMessage, BLOCK_DATA_SIZE * pSizeInfo1->mMessage,
        outLangBlock->mMessage, theNumTracks, theIs2byte);
    SeparateInfo(pISRC, BLOCK_DATA_SIZE * pSizeInfo2->mMCNorISRC,
        outLangBlock->mMCNorISRC, theNumTracks, false);


    free(pIdentifier);
    free(pGenre);
    free(pTitle);
    free(pPerformer);
    free(pSongwriter);
    free(pComposer);
    free(pArranger);
    free(pMessage);
    free(pISRC);

    return true;
}
//---------------------------------------------------------------------------

//obt@gbNPʂɕ
void SeparateInfo(
    char *inBuffer,
    int inBufSize,
    void* outTracks,
    int inNumTracks,
    bool inIs2byteChar)
{
    if(!inBuffer || !outTracks || !inBufSize) return;
    int i;

    int thePos = 0;
    if(inIs2byteChar)
        for(i=0;i<=inNumTracks;i++){

            short* pBuf = (short*)inBuffer;
            int j;
            for(j=0; j + thePos < inBufSize / 2 ;j++){
                if(pBuf[thePos + j]==0) break;
            }

            memcpy((char*)outTracks + MAX_TEXT * i, pBuf + thePos, min(j * 2, MAX_TEXT - 2));

            thePos += j;
            thePos ++;
        }
    else
        for(i=0;i<=inNumTracks;i++){

            char* pBuf=(char*)inBuffer;
            int j;
            for(j=0; j + thePos < inBufSize ;j++){
                if(pBuf[thePos + j]==0) break;
            }

            memcpy((char*)outTracks + MAX_TEXT * i, pBuf + thePos, min(j, MAX_TEXT - 1));

            thePos += j;
            thePos ++;
        }
}
//---------------------------------------------------------------------------

//R[h2oCggp邩ۂԂ
bool Is2ByteCharset(CdTextCharsets inCharset)
{
    switch (inCharset){
    case ccISO8859:
        return false;
    case ccASCII:
        return false;
    case ccMS_JIS:
        return true;
    default:
        return false;
    }
}
//---------------------------------------------------------------------------
